#include "hellosphere.h"
#include <iostream>

#define deg2rad(x) float(M_PI)*(x)/180.f

// CTOR
SimpleSphere::SimpleSphere(int width, int height, MainWindow* w) : Scene(width, height, w), _activecamera(0), _camera(nullptr) {
    // Initialize last shaders
    _lastVertex   = _w->vertexShader;
    _lastFragment = _w->fragmentShader;

    // Connect programm generation slot ?

    // Compute the shape to draw -> Menu ?
    //drawCube(0.1f);
    //drawUVSphere(50, _radius);
    drawIcoSphere(6, _radius);

    std::cout << "Vertices count: " << _vertices.size() / 3 << std::endl;
    std::cout << "Edges count: " << _indices.size() << std::endl;
    std::cout << "Triangles count: " << _indices.size() / 3 << std::endl;

    // 1. Generate geometry buffers
    glGenBuffers(1, &_vbo) ;
    glGenBuffers(1, &_nbo) ;
    glGenBuffers(1, &_ebo) ;
    glGenVertexArrays(1, &_vao) ;
    // 2. Bind Vertex Array Object
    glBindVertexArray(_vao);
        // 3. Copy our vertices array in a buffer for OpenGL to use
        glBindBuffer(GL_ARRAY_BUFFER, _vbo);
        glBufferData(GL_ARRAY_BUFFER, _vertices.size()*sizeof (GLfloat), _vertices.data(), GL_STATIC_DRAW);
        // 4. Then set our vertex attributes pointers
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(GLfloat), (GLvoid*)0);
        glEnableVertexAttribArray(0);
        // 5. Copy our normals array in a buffer for OpenGL to use
        glBindBuffer(GL_ARRAY_BUFFER, _nbo);
        glBufferData(GL_ARRAY_BUFFER, _normals.size()*sizeof (GLfloat), _normals.data(), GL_STATIC_DRAW);
        // 6. Copy our vertices array in a buffer for OpenGL to use
        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(GLfloat), (GLvoid*)0);
        glEnableVertexAttribArray(1);
        // 7. Copy our index array in a element buffer for OpenGL to use
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, _ebo);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, _indices.size()*sizeof (GLfloat), _indices.data(), GL_STATIC_DRAW);
    //6. Unbind the VAO
    glBindVertexArray(0);

    // Camera Initialization
    _cameraselector.push_back( []()->Camera*{return new EulerCamera(glm::vec3(0.f, 0.f, 1.f));} );
    _cameraselector.push_back( []()->Camera*{return new TrackballCamera(glm::vec3(0.f, 0.f, 1.f),glm::vec3(0.f, 1.f, 0.f),glm::vec3(0.f, 0.f, 0.f));} );

    _camera.reset(_cameraselector[_activecamera]());

    _camera->setviewport(glm::vec4(0.f, 0.f, _width, _height));
    _view = _camera->viewmatrix();

    _projection = glm::perspective(_camera->zoom(), float(_width) / _height, 0.1f, 100.0f);

    generateProgram();
    glUseProgram(_program);
}

SimpleSphere::~SimpleSphere() {}

void SimpleSphere::resize(int width, int height){
    Scene::resize(width, height);
    _camera->setviewport(glm::vec4(0.f, 0.f, _width, _height));
}

void SimpleSphere::draw() {
    // Set OpenGL Render Mode
    Scene::draw();

    // Recompute program
    if(_lastVertex != _w->vertexShader || _lastFragment != _w->fragmentShader) {
        generateProgram();
        glUseProgram(_program);
    }

    // Uniforms
    glUniform3f(glGetUniformLocation(_program, "center"), 0.0f, 0.0f, 0.0f);
    glUniform1f(glGetUniformLocation(_program, "radius"), _radius);
    glUniform1f(glGetUniformLocation(_program, "errorMax"), _errorMax);
    glUniform3f(glGetUniformLocation(_program, "cameraView"), _camera->_front.x, _camera->_front.y, _camera->_front.z);

    // Camera
    _view = _camera->viewmatrix();

    glUniformMatrix4fv( glGetUniformLocation(_program, "model"), 1, GL_FALSE, glm::value_ptr(_model));
    glUniformMatrix4fv( glGetUniformLocation(_program, "view"), 1, GL_FALSE, glm::value_ptr(_view));
    glUniformMatrix4fv( glGetUniformLocation(_program, "projection"), 1, GL_FALSE, glm::value_ptr(_projection));

    // Draw the shape
    glBindVertexArray(_vao);
    glDrawElements(GL_TRIANGLES, _indices.size(), GL_UNSIGNED_INT, 0);
    glBindVertexArray(0);
}

void SimpleSphere::mouseclick(int button, float xpos, float ypos) {
    _button = button;
    _mousex = xpos;
    _mousey = ypos;
    _camera->processmouseclick(_button, xpos, ypos);
}

void SimpleSphere::mousemove(float xpos, float ypos) {
    _camera->processmousemovement(_button, xpos, ypos, true);
}

void SimpleSphere::keyboardmove(int key, double time) {
    _camera->processkeyboard(Camera_Movement(key), time);
}

bool SimpleSphere::keyboard(unsigned char k) {
    switch(k) {
        case 'p':
            _activecamera = (_activecamera+1)%2;
            _camera.reset(_cameraselector[_activecamera]());
            _camera->setviewport(glm::vec4(0.f, 0.f, _width, _height));
            return true;
        default:
            return false;
    }
}

void SimpleSphere::generateProgram() {
    // Change last shader variables
    _lastVertex = _w->vertexShader;
    _lastFragment = _w->fragmentShader;

    // Initialize shaders
    GLint success;
    GLchar infoLog[512]; // warning fixed size ... request for LOG_LENGTH!!!
    GLuint vertexshader, fragmentshader;

    /* VERTEX SHADER */
    // 1. Generate the shader
    vertexshader = glCreateShader(GL_VERTEX_SHADER);
    // 2. set the source
    auto pointerVertex = _w->vertexShaders[_w->vertexShader].c_str();
    glShaderSource(vertexshader, 1, &pointerVertex, NULL);
    // 3. Compile
    glCompileShader(vertexshader);
    // 4. test for compile error
    glGetShaderiv(vertexshader, GL_COMPILE_STATUS, &success);
    if(!success) {
        glGetShaderInfoLog(vertexshader, 512, NULL, infoLog);
        std::cerr << "ERROR::SHADER::VERTEX::COMPILATION_FAILED\n" << infoLog << std::endl;
    }

    /* FRAGMENT SHADER */
    fragmentshader = glCreateShader(GL_FRAGMENT_SHADER);
    auto pointerFragment = _w->fragmentShaders[_w->fragmentShader].c_str();
    glShaderSource(fragmentshader, 1, &pointerFragment, NULL);
    glCompileShader(fragmentshader);
    glGetShaderiv(fragmentshader, GL_COMPILE_STATUS, &success);
    if(!success) {
        glGetShaderInfoLog(fragmentshader, 512, NULL, infoLog);
        std::cerr << "ERROR::SHADER::FRAGMENT::COMPILATION_FAILED\n" << infoLog << std::endl;
    }

    /* Program Object */
    // 1. Generate the program
    _program = glCreateProgram();
    // 2. Attach the shaders to the program
    glAttachShader(_program, vertexshader);
    glAttachShader(_program, fragmentshader);
    // 3. Link the program
    glLinkProgram(_program);
    // 4. Test for link errors
    glGetProgramiv(_program, GL_LINK_STATUS, &success);
    if(!success) {
        glGetProgramInfoLog(_program, 512, NULL, infoLog);
        std::cerr << "ERROR::SHADER::LINK_FAILED\n" << infoLog << std::endl;
    }
    glDeleteShader(vertexshader);
    glDeleteShader(fragmentshader);
}

void SimpleSphere::drawCube(float size) {
    // Initialise geometric data
    _vertices = {
        -size, -size, -size,
        size, -size, -size,
        size, size, -size,
        -size, size, -size,

        -size, -size, size,
        size, -size, size,
        size, size, size,
        -size, size, size,
    };

    _normals = {
        -size, -size, -size,
        size, -size, -size,
        size, size, -size,
        -size, size, -size,

        -size, -size, size,
        size, -size, size,
        size, size, size,
        -size, size, size,
    };

    _indices = {
        // Note that we start from 0!
        0, 1, 2,   // First Triangle
        0, 2, 3,   // Second Triangle

        1, 5, 6,
        1, 6, 2,

        4, 0, 3,
        4, 3, 7,

        5, 4, 7,
        5, 7, 6,

        0, 1, 5,
        0, 5, 4,

        3, 2, 6,
        3, 6, 7
    };

    return;
}

void SimpleSphere::drawUVSphere(int resolution, float radius) {
    // Rubans et méridiens
    for(int u = 0; u < resolution; ++u) {
        float theta = float(u) / float(resolution) * 2.f * M_PI;
        for (int v = 1; v < resolution; ++v) {
            float phi = (float(v) / float(resolution) * M_PI) - M_PI_2;
            /* Vertices */
            // Compute coordonates
            float x = radius * cos(theta) * cos(phi);
            float y = radius * sin(phi);
            float z = radius * sin(theta) * cos(phi);

            // Add to vector
            _vertices.push_back(x);
            _vertices.push_back(y);
            _vertices.push_back(z);

            // Normals
            glm::vec3 vertex = glm::vec3(x, y, z);
            glm::vec3 normal = glm::normalize(vertex);
            _normals.push_back(normal.x);
            _normals.push_back(normal.y);
            _normals.push_back(normal.z);
        }
    }

    // Indices
    for(int u = 0; u < resolution; ++u) {
        for(int v = 0; v < resolution - 2; ++v) {
            int sommetA = u * (resolution - 1) + v;
            int sommetB = ((u+1) % resolution) * (resolution - 1) + v;
            int sommetC = (u    * (resolution - 1) + v + 1);
            int sommetD = ((u+1) % resolution) * (resolution - 1) + v + 1;

            // Triangle 1
            _indices.push_back(sommetA);
            _indices.push_back(sommetD);
            _indices.push_back(sommetB);
            // Triangle 2
            _indices.push_back(sommetA);
            _indices.push_back(sommetC);
            _indices.push_back(sommetD);
        }
    }

    /* Pôles */
    for(int i = 0; i < 2; i++) {
        // Vertices
        float x = 0.f;
        float y = round(i - 0.5) * radius;
        float z = 0.f;
        _vertices.push_back(x);
        _vertices.push_back(y);
        _vertices.push_back(z);

        // Normals
        glm::vec3 vertex = glm::vec3(x, y, z);
        glm::vec3 normal = glm::normalize(vertex);
        _normals.push_back(normal[0]);
        _normals.push_back(normal[1]);
        _normals.push_back(normal[2]);

        // Indices
        for(int j = 0; j < resolution; ++j) {
            unsigned int sommetA = _vertices.size() / 3 - 1;
            unsigned int sommetB = i * (resolution-2) + j * (resolution - 1);
            unsigned int sommetC = i * (resolution-2) + ((j + 1) % resolution) * (resolution - 1);
            if (i)
                _indices.insert(_indices.end(), {sommetA, sommetC, sommetB});
            else
                _indices.insert(_indices.end(), {sommetA, sommetB, sommetC});
        }
    }

    // Erreur max pour le shader de différence
    // L'erreur max se situe au centre d'un quadrilatère sur les bandes de l'équateur
    // Ce centre équivaut au centre de l'arête qui consitue la diagonale de ce quadrilatère
    uint indexSommetA = (_indices.at(_indices.size() / 2) + 1) * 3;
    uint indexSommetB = (_indices.at(_indices.size() / 2) + 2) * 3;
    glm::vec3 sommetA = glm::vec3(_vertices.at(indexSommetA), _vertices.at(indexSommetA+1),_vertices.at(indexSommetA+2));
    glm::vec3 sommetB = glm::vec3(_vertices.at(indexSommetB), _vertices.at(indexSommetB+1),_vertices.at(indexSommetB+2));
    glm::vec3 centre = (sommetA + sommetB) / 2.0f;
    _errorMax = abs( glm::length(centre) - _radius) / _radius;
}

void SimpleSphere::drawIcoSphere(int resolution, float radius) {
    /* Constantes */
    float phi = (1.f + sqrt(5.f)) / 2.f;
    // r = a/2 * sqrt(2+phi)
    // => a = 2*r / sqrt(2+phi)
    // Or constante = a/2 donc a = r / sqrt(2+phi)
    float a = radius / sqrt(2.f + phi);


    /* Vertices */
    // Rectangle 1
    for(int i = 0; i < 4; i++) {
        float x = round((i/2)%2 - 0.5f) * a * phi;
        float y = round((i%2) - 0.5f) * a;
        float z = 0.f;
        _vertices.push_back(x);
        _vertices.push_back(y);
        _vertices.push_back(z);

        glm::vec3 vertex = glm::vec3(x, y ,z);
        glm::vec3 normal = glm::normalize(vertex);
        _normals.push_back(normal[0]);
        _normals.push_back(normal[1]);
        _normals.push_back(normal[2]);
    }

    // Rectangle 2 (1, 0, a)
    for(int i = 0; i < 4; i++) {
        float x = round((i%2) - 0.5f) * a;
        float y = 0.f;
        float z = round((i/2)%2 - 0.5f) * a * phi;
        _vertices.push_back(x);
        _vertices.push_back(y);
        _vertices.push_back(z);

        glm::vec3 vertex = glm::vec3(x, y, z);
        glm::vec3 normal = glm::normalize(vertex);
        _normals.push_back(normal[0]);
        _normals.push_back(normal[1]);
        _normals.push_back(normal[2]);
    }

    // Rectangle 3 (0, phi, 1)
    for(int i = 0; i < 4; i++) {
        float x = 0.f;
        float y = round((i/2)%2 - 0.5f) * a * phi;
        float z = round((i%2) - 0.5f) * a;
        _vertices.push_back(x);
        _vertices.push_back(y);
        _vertices.push_back(z);

        glm::vec3 vertex = glm::vec3(x, y, z);
        glm::vec3 normal = glm::normalize(vertex);
        _normals.push_back(normal[0]);
        _normals.push_back(normal[1]);
        _normals.push_back(normal[2]);
    }

    /* Indices */
    _indices.insert(_indices.end(), {  6,  7, 11});
    _indices.insert(_indices.end(), {  6,  9,  7});
    _indices.insert(_indices.end(), {  7,  3, 11});
    _indices.insert(_indices.end(), {  7,  9,  2});
    _indices.insert(_indices.end(), {  7,  2,  3});
    _indices.insert(_indices.end(), {  5,  3,  2});
    _indices.insert(_indices.end(), {  5, 10,  3});
    _indices.insert(_indices.end(), {  5,  2,  8});
    _indices.insert(_indices.end(), {  5,  4, 10});
    _indices.insert(_indices.end(), {  5,  8,  4});
    _indices.insert(_indices.end(), {  4,  1, 10});
    _indices.insert(_indices.end(), {  4,  8,  0});
    _indices.insert(_indices.end(), {  4,  0,  1});
    _indices.insert(_indices.end(), {  6,  1,  0});
    _indices.insert(_indices.end(), {  1,  6, 11});
    _indices.insert(_indices.end(), {  0,  9,  6});
    _indices.insert(_indices.end(), {  9,  0,  8});
    _indices.insert(_indices.end(), {  9,  8,  2});
    _indices.insert(_indices.end(), { 10,  1, 11});
    _indices.insert(_indices.end(), { 10, 11,  3});

    /* Subdivision */
    for (int i = 1; i < resolution; ++i) {
        std::vector<GLuint> _newIndices;
        // 1. Calculer les points entre chaque arête
        // Stockés dans une map dont la clé est la paire d'arêtes, et la valeur l'indice du sommet
        std::map<std::pair<int, int>, uint> mapAreteSommet;

        // Parcours des triangles
        for (unsigned int j = 0; j < _indices.size(); j+=3) {
            uint indexSommetA = _indices[j];
            uint indexSommetB = _indices[j+1];
            uint indexSommetC = _indices[j+2];
            uint indexSommetAB, indexSommetBC, indexSommetCA;

            // Calcul du point entre le sommet A et le sommet B
            // Récupération de l'indice existant pour le point éventuellement déjà calculé
            if(mapAreteSommet.count(std::pair<int, int> {indexSommetA, indexSommetB}) != 0)
                indexSommetAB = mapAreteSommet.at(std::pair<int, int> {indexSommetA, indexSommetB});
            else if(mapAreteSommet.count(std::pair<int, int> {indexSommetB, indexSommetA}) != 0)
                indexSommetAB = mapAreteSommet.at(std::pair<int, int> {indexSommetB, indexSommetA});
            // Sinon création du point
            else {
                // Calcul de la coordonnée du point à la moité de l'arête
                glm::vec3 sommetA = glm::vec3(_vertices[indexSommetA*3], _vertices[indexSommetA*3+1], _vertices[indexSommetA*3+2]);
                glm::vec3 sommetB = glm::vec3(_vertices[indexSommetB*3], _vertices[indexSommetB*3+1], _vertices[indexSommetB*3+2]);

                glm::vec3 sommetAB = (sommetA + sommetB) / 2.0f;

                // Repositionnement du point sur la sphère hypothétique
                float vertexDistance = glm::length(sommetAB);
                sommetAB *= radius / vertexDistance;

                // Insertion dans les vector _vertices et _indices
                _vertices.insert(_vertices.end(), {sommetAB.x, sommetAB.y, sommetAB.z});
                _normals.insert(_normals.end(), {sommetAB.x, sommetAB.y, sommetAB.z});

                indexSommetAB = _vertices.size() / 3 - 1;
                mapAreteSommet[std::pair<int, int> {indexSommetA, indexSommetB}] = indexSommetAB;
            }

            // Calcul du point entre le sommet B et le sommet C
            // Récupération de l'indice existant pour le point éventuellement déjà calculé
            if(mapAreteSommet.count(std::pair<int, int> {indexSommetB, indexSommetC}) != 0)
                indexSommetBC = mapAreteSommet.at(std::pair<int, int> {indexSommetB, indexSommetC});
            else if(mapAreteSommet.count(std::pair<int, int> {indexSommetC, indexSommetB}) != 0)
                indexSommetBC = mapAreteSommet.at(std::pair<int, int> {indexSommetC, indexSommetB});
            // Sinon création du point
            else {
                // Calcul de la coordonnée du point à la moité de l'arête
                glm::vec3 sommetB = glm::vec3(_vertices[indexSommetB*3], _vertices[indexSommetB*3+1], _vertices[indexSommetB*3+2]);
                glm::vec3 sommetC = glm::vec3(_vertices[indexSommetC*3], _vertices[indexSommetC*3+1], _vertices[indexSommetC*3+2]);

                glm::vec3 sommetBC = (sommetB + sommetC) / 2.0f;

                // Repositionnement du point sur la sphère hypothétique
                float vertexDistance = glm::length(sommetBC);
                sommetBC *= radius / vertexDistance;

                // Insertion dans les vector _vertices et _indices
                _vertices.insert(_vertices.end(), {sommetBC.x, sommetBC.y, sommetBC.z});
                _normals.insert(_normals.end(), {sommetBC.x, sommetBC.y, sommetBC.z});

                indexSommetBC = _vertices.size() / 3 - 1;
                mapAreteSommet[std::pair<int, int> {indexSommetB, indexSommetC}] = indexSommetBC;
            }

            // Calcul du point entre le sommet C et le sommet A
            // Récupération de l'indice existant pour le point éventuellement déjà calculé
            if(mapAreteSommet.count(std::pair<int, int> {indexSommetC, indexSommetA}) != 0)
                indexSommetCA = mapAreteSommet.at(std::pair<int, int> {indexSommetC, indexSommetA});
            else if(mapAreteSommet.count(std::pair<int, int> {indexSommetA, indexSommetC}) != 0)
                indexSommetCA = mapAreteSommet.at(std::pair<int, int> {indexSommetA, indexSommetC});
            // Sinon création du point
            else {
                glm::vec3 sommetC = glm::vec3(_vertices[indexSommetC*3], _vertices[indexSommetC*3+1], _vertices[indexSommetC*3+2]);
                glm::vec3 sommetA = glm::vec3(_vertices[indexSommetA*3], _vertices[indexSommetA*3+1], _vertices[indexSommetA*3+2]);
                // Calcul de la coordonnée du point à la moité de l'arête

                glm::vec3 sommetCA = (sommetC + sommetA) / 2.0f;

                // Repositionnement du point sur la sphère hypothétique
                float vertexDistance = glm::length(sommetCA);
                sommetCA *= radius / vertexDistance;

                // Insertion dans les vector _vertices et _indices
                _vertices.insert(_vertices.end(), {sommetCA.x, sommetCA.y, sommetCA.z});
                _normals.insert(_normals.end(), {sommetCA.x, sommetCA.y, sommetCA.z});

                indexSommetCA = _vertices.size() / 3 - 1;
                mapAreteSommet[std::pair<int, int> {indexSommetC, indexSommetA}] = indexSommetCA;
            }

            // Création des nouveaux triangles
            _newIndices.insert(_newIndices.end(), { indexSommetA,  indexSommetAB, indexSommetCA});
            _newIndices.insert(_newIndices.end(), {indexSommetAB,   indexSommetB, indexSommetBC});
            _newIndices.insert(_newIndices.end(), {indexSommetBC,   indexSommetC, indexSommetCA});
            _newIndices.insert(_newIndices.end(), {indexSommetAB,  indexSommetBC, indexSommetCA});
        }
        // Remplacer le tableau d'incides
        _indices.swap(_newIndices);
    }

    // Erreur max pour le shader de différence
    // L'erreur max se situe au centre d'un triangle de l'icosaèdre
    uint indexSommetA = _indices.at(0) * 3;
    uint indexSommetB = _indices.at(1) * 3;
    uint indexSommetC = _indices.at(2) * 3;
    glm::vec3 sommetA = glm::vec3(_vertices.at(indexSommetA), _vertices.at(indexSommetA+1),_vertices.at(indexSommetA+2));
    glm::vec3 sommetB = glm::vec3(_vertices.at(indexSommetB), _vertices.at(indexSommetB+1),_vertices.at(indexSommetB+2));
    glm::vec3 sommetC = glm::vec3(_vertices.at(indexSommetC), _vertices.at(indexSommetC+1),_vertices.at(indexSommetC+2));
    glm::vec3 centre = (sommetA + sommetB + sommetC) / 3.0f;
    _errorMax = abs( glm::length(centre) - _radius) / _radius;
}
